/*

This file is (C) Copyright 2002 Erik S. Johansen <einride@einride.org>, all 
rights reserved.

If you work for or represent an organisation, agency or corporation that is 
wholly or partially funded by a government -OR- you are in any way associated 
with the company Cyber Operations, the following licence applies.

You may not use this file or any derivatives of it for any purpose whatsoever,
including reading beyond this comment, redistribution or compiling, without 
explicit written permission from the copyright holder. 

If you do not work for or represent an organisation, agency or corporation 
that is wholly or partially funded by a government -AND- you are not in any 
way associated with the company Cyber Operations, the GNU General Public 
License, as described in the accompanying file GPL, applies.

*/
#include "ipc.h"
#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <sys/un.h>
#include <sys/time.h>
#include <unistd.h>
#include <error.h>
#include <errno.h>
#include <fcntl.h>

#define IPCT_UNIXSOCKET 1

#define IPCF_CREATED    1
#define IPCF_ATTACHED   2

void mktempfilename(char * buf, int len) {  
  int i;
  for (i=0;i<len;i++) 
    buf[i] = ((rand() % 2) * 32) + (rand() % 26) + 65;
  buf[len] = 0;
}

ipc_struct * ipc_create_unixsocket() {
  ipc_struct * ipc = (ipc_struct *) malloc(sizeof(ipc_struct));
  struct sockaddr_un sa;
  memset(ipc, 0, sizeof(ipc_struct));
  ipc->type = IPCT_UNIXSOCKET;
  ipc->flags = IPCF_CREATED;
  ipc->fd2 = socket(AF_UNIX, SOCK_STREAM, 0);
  ipc->fd1 = -1;
  if (ipc->fd2<0) {
    free(ipc);
    return 0;
  }
  strcpy(sa.sun_path, "/tmp/");
  mktempfilename(sa.sun_path + strlen(sa.sun_path), 58);
  sa.sun_family = AF_UNIX;
  if (bind(ipc->fd2, (struct sockaddr *) &sa, sizeof(struct sockaddr_un))) {
    close(ipc->fd2);
    free(ipc);
    return 0;
  }
  if (chmod(sa.sun_path, S_IRUSR | S_IWUSR | S_IRGRP | S_IWGRP | S_IROTH | S_IWOTH)) {
    close(ipc->fd2);
    unlink(sa.sun_path);
    free(ipc);
    return 0;
  }
  sprintf(ipc->id, "%i:%s", ipc->type, sa.sun_path);
  if (listen(ipc->fd2, 1)) {
    close(ipc->fd2);
    unlink(sa.sun_path);
    free(ipc);
    return 0;
  }
  if (fcntl(ipc->fd2, F_SETFL, O_NONBLOCK)) {
    close(ipc->fd2);
    unlink(sa.sun_path);
    free(ipc);
    return 0;
  }
  return ipc;
}

ipc_struct * ipc_create() {
  return ipc_create_unixsocket();
}

ipc_struct * ipc_attach_unixsocket(char * name) {
  ipc_struct * ipc = (ipc_struct *) malloc(sizeof(ipc_struct));
  struct sockaddr_un sa;
  memset(ipc, 0, sizeof(ipc_struct));
  ipc->fd2 = -1;
  ipc->type = IPCT_UNIXSOCKET;
  ipc->flags = IPCF_ATTACHED;
  ipc->fd1 = socket(AF_UNIX, SOCK_STREAM, 0);
  if (ipc->fd1<0) {
    free(ipc);
    return 0;
  }
  sa.sun_family = AF_UNIX;
  strcpy(sa.sun_path, name);
  if (connect(ipc->fd1, (struct sockaddr *) &sa, sizeof(sa))) {
    free(ipc);
    return 0;
  }
  if (fcntl(ipc->fd1, F_SETFL, O_NONBLOCK)) {
    close(ipc->fd1);
    unlink(sa.sun_path);
    free(ipc);
    return 0;
  }
  sprintf(ipc->id, "%i:%s", ipc->type, sa.sun_path);
  return ipc;
}

ipc_struct * ipc_attach(char * id) {
  int iType;
  char sSpec[256];
  if (sscanf(id, "%i:%s", &iType, sSpec)!=2) {
    return 0;
  }
  switch(iType) {
  case IPCT_UNIXSOCKET:
    return ipc_attach_unixsocket(sSpec);
  default:
    return 0;
  }
}

void ipc_close_unixsocket(ipc_struct * ipc) {
  int i;
  char sockname[256];
  if (ipc->fd1>=0)
    close(ipc->fd1);
  if (ipc->fd2>=0)
    close(ipc->fd2);
  sscanf(ipc->id, "%i:%s", &i, sockname);
  unlink(sockname);
  if (ipc->intbuf)
    free(ipc->intbuf);
  free(ipc);
}

void ipc_close(ipc_struct * ipc) {
  switch (ipc->type) {
  case IPCT_UNIXSOCKET:
    ipc_close_unixsocket(ipc);
    break;
  default:
    break;
  }
}

int ipc_read_unixsocket(ipc_struct * ipc, char * buf, int bufsize) {
  int res;
  unsigned int ssa;
  struct sockaddr_un sa;
  struct timeval tv = {0, 0};
  fd_set rfds, efds;
  memset(&sa, 0, sizeof(sa));
  if (ipc->fd1 < 0) {
    ssa = sizeof(sa);
    ipc->fd1 = accept(ipc->fd2, (struct sockaddr *) &sa, &ssa);
    if (ipc->fd1 < 0) {
      if (errno==EAGAIN)
	return 0;
      else
	return -1;
    }
    fcntl(ipc->fd1, F_SETFL, O_NONBLOCK);
  }
  FD_ZERO(&rfds);
  FD_ZERO(&efds);
  FD_SET(ipc->fd1, &rfds);
  FD_SET(ipc->fd1, &efds);
  if (!(res = select(ipc->fd1 + 1, &rfds, 0, &efds, &tv)))
    return 0;
  if (FD_ISSET(ipc->fd1, &rfds)) {
    res = read(ipc->fd1, buf, bufsize);
    if (res<=0) 
      return -1;
    else
      return res;
  } else if (FD_ISSET(ipc->fd1, &efds)) {
    return -1;
  }
  // Shouldn't happen....
  return -1;
}

int ipc_read(ipc_struct * ipc, char * buf, int bufsize) {
  switch (ipc->type) {
  case IPCT_UNIXSOCKET:
    return ipc_read_unixsocket(ipc, buf, bufsize);
  default:
    return 0;
  }
}

int ipc_write_unixsocket(ipc_struct * ipc, char * buf, int bufsize) {
  int res;
  if (ipc->fd1 < 0) {
    ipc->fd1 = accept(ipc->fd2, 0, 0);
    if (ipc->fd1 < 0) {
      if (errno==EAGAIN)
	return 0;
      else
	return -1;
    }
  }
  res = write(ipc->fd1, buf, bufsize);
  if ((res<0) && (errno==EAGAIN))
    return 0;
  else if (!res) {
    errno = ECONNRESET;
    return -1;
  }
  return res;
}

int ipc_write(ipc_struct * ipc, char * buf, int bufsize) {
  switch (ipc->type) {
  case IPCT_UNIXSOCKET:
    return ipc_write_unixsocket(ipc, buf, bufsize);
  default:
    return 0;
  }
}

int ipc_sendblock(ipc_struct * ipc, void * data, int datasize) {
  int res=ipc_write(ipc, (char *) &datasize, 4);
  if (res<=0) 
    return res;
  else if (res != 4)
    return -1;
  res=ipc_write(ipc, (char *) data, datasize);
  if (res<=0)
    return res;
  else if (res != datasize)
    return -1;
  return 1;
}

int ipc_readblock(ipc_struct * ipc, void * data, int * datasize) {
  int res, bsize, readfault=0;
  char tmp[1024];
  do {
    res=ipc_read(ipc, (void *) &tmp, sizeof(tmp));
    if (res>0) {
      if (res + ipc->intbufp > ipc->intbufs) {
	ipc->intbufs += sizeof(tmp);
	ipc->intbuf = (char *) realloc(ipc->intbuf, ipc->intbufs);
      }
      memcpy(&ipc->intbuf[ipc->intbufp], &tmp, res);
      ipc->intbufp += res;
    } else if (res<0) {
      readfault = 1;
      break;
    }
  } while (res);
  
  if (!datasize)
    return 0;
  if (ipc->intbufp<4) {
    *datasize = 0;
    if (readfault)
      return -1;
    else
      return 0;
  }
  bsize = *(int *) ipc->intbuf;
  if (ipc->intbufp < bsize + 4) {
    *datasize = 0;
    if (readfault)
      return -1;
    else
      return 0;
  }
  if (*datasize < bsize) {
    *datasize = bsize;
    return 0;
  }
  memcpy(data, &ipc->intbuf[4], bsize);
  if (ipc->intbufp > bsize + 4) 
    memmove(ipc->intbuf, &ipc->intbuf[4+bsize], (ipc->intbufp - (bsize +4)));
  ipc->intbufp -= (bsize + 4);
  *datasize = bsize;
  return 1;
}
