/*

This file is (C) Copyright 2002 Erik S. Johansen <einride@einride.org>, all 
rights reserved.

If you work for or represent an organisation, agency or corporation that is 
wholly or partially funded by a government -OR- you are in any way associated 
with the company Cyber Operations, the following licence applies.

You may not use this file or any derivatives of it for any purpose whatsoever,
including reading beyond this comment, redistribution or compiling, without 
explicit written permission from the copyright holder. 

If you do not work for or represent an organisation, agency or corporation 
that is wholly or partially funded by a government -AND- you are not in any 
way associated with the company Cyber Operations, the GNU General Public 
License, as described in the accompanying file GPL, applies.

*/
#include "main.h"
#include "options.h"
#include "inject.h"
#include "util.h"
#include "child.h"
#include "ui.h"
#include "debug.h"
#include <time.h>

extern struct app_arg_struct args;
extern struct app_glob_struct glob;
extern struct child_glob_struct cglob;
extern char * binname;

/*************************************************************************
 * Misc "utility" functions
 *************************************************************************/
extern struct option optlist[];

/*
  Always: any <0> in the line is replaced with ".", and a 0 is appended to
          the line
  GUI: Appends the string to data window
  STD: Writes the string to stdout and flushes
*/
void out_data(char * line, int len) {
  int i;
  for (i=0;i<len;i++)
    if (!line[i])
      line[i]='.';
  line[len]=0;
  if (args.gui) {
    wprintw(glob.datawin, "%s", line);
    wrefresh(glob.datawin);
  } else {
    printf("%s", line);
  }
}

/* Add data to tasks wdata buffer, and dump it to stdout if we're in interactive mode */
void append_task_data(task * task, char * data, int datalen, int direction) {
  if (task->fd) {
    util_cyclicbufferappend(&task->wdata[0], sizeof(task->wdata), &task->wdatastart, &task->wdataend, data, datalen);

    if ((glob.commandmode==MODE_INTERACTIVE) && (glob.currentfd == task->fd))
      ui_appenddata(data, datalen);
  } else {
    if ((direction==DIR_OUT) || glob.echoinput) {
      util_cyclicbufferappend(&task->wdata[0], sizeof(task->wdata), &task->wdatastart, &task->wdataend, data, datalen);
      if ((glob.commandmode==MODE_INTERACTIVE) && (glob.currentfd == 0))
	ui_appenddata(data, datalen);
    }
  }
}

int sanitycheckargs() {
  if (!args.pid)
    return 0;
  if (!args.tasks && !args.expire) {
    ui_showmessage(UI_LEVEL_NORMAL, "No continous tasks - setting expiry to 10 seconds");
    args.expire = 10;
  }
  return 1;
}

/*************************************************************************
 * redirect/restore funcs
 *************************************************************************/
static short error_count = 0;

fdrelay * redirectfd(int fd) {
  fdrelay * fr;
  int trgSock, fdflags=0xFFFFFFFF, i;

  for (i=0;i<glob.fdcount;i++) {
    if (glob.fdlist[i].fd==fd) {
      fdflags = glob.fdlist[i].flags;
      break;
    }
  }
  if (fdflags==0xFFFFFFFF) {
    if (error_count > 4) {
      ui_showmessage(UI_LEVEL_ERROR, "To many redirection errors for %i exiting..", fd);
      exit(EXIT_FAILURE);
    }
    ui_showmessage(UI_LEVEL_ERROR, "Invalid FD %i specified for redirection", fd);
    error_count++;
    return 0;
  }
  fr = util_addrelay();
  if (!fr) 
    return 0;

  /* Make target create an unix socket and connect to us */
  if (inject_forceconnect((char *) &fr->ipc->id[2], &trgSock)) {
    util_delrelay(fr);
    kill(args.pid, SIGCHLD);
    return 0;
  }
  
  /* Make target fcntl the socket so it matches the fd */
  if (inject_forcefcntlsetfl(fd, fdflags)) {
    util_delrelay(fr);
    kill(args.pid, SIGCHLD);
    return 0;
  }

  /* Make target dup fd, close fd, dup socket to orig fd, close socket */
  if (inject_forceredirectfd(fd, trgSock, &fr->trg_SavedFD)) {
    util_delrelay(fr);
    kill(args.pid, SIGCHLD);
    return 0;
  }
  fr->trg_FD = fd;
  //  glob.blockstd = 1;
  return fr;
}

int restorefd(int fd) {
  fdrelay * fr = util_findrelay(fd);
  if (!fr)
    return 1;
  if (glob.process->flags & TRACE_EXITED)
    return 0;
  if (inject_forcerestorefd(fr->trg_SavedFD, fr->trg_FD)) {
    ui_showmessage(UI_LEVEL_NORMAL, "Error restoring targets fd %i", fr->trg_FD);
    kill(args.pid, SIGCHLD);
    return 1;
  }
  util_delrelay(fr);
  // glob.blockstd = 1;
  return 0;
}


/* Setup the target, force a child and get fdlist */
int prepareTarget() {
  char * buf = 0, *p;
  int done = 0, bufsize = 0, i;
  mapentry * bmap, * cmap;
  /* Try to find a point for code insertion */

  /* First get the targets commandline, and then get the bin name */
  if (!util_getprocessbinname(args.pid, 1, glob.binname, sizeof(glob.binname)))
    glob.binname[0] = 0;
  
  /* Now read memory maps and scan through them looking for a good area
     It's required to be readable, executable and not shared, and should
     preferrably be in the file with a name matching binname */
  glob.map = util_makemapentry(args.pid);
  bmap = 0;
  if (glob.map) {
    cmap = glob.map;
    while (cmap) {
      if (cmap->vm_read && cmap->vm_exec && !cmap->vm_shared) {
	p = strrchr(cmap->file, '/');
	if (!p)
	  p=cmap->file;
	else
	  p++;
	if (!strncmp(p, glob.binname, strlen(glob.binname))) {
	  // Got the bin itself, use that
	  bmap = cmap;
	  break;
	} else if (!strstr(p, ".so")) {
	  if (!bmap)
	    bmap = cmap;
	}
      }
      cmap = cmap->next;
    }
  }
  if (bmap) {
    glob.insertionpoint = bmap->vm_start + 0x400;
  } else {
    ui_showmessage(UI_LEVEL_NORMAL, "Can''t find a vm block suitable for code injection");
    return 1;
  }

  /* Attach as debugger of target */
  glob.process = trace_attach(args.pid);
  if (!glob.process) {
    ui_showmessage(UI_LEVEL_NORMAL, "Error %i occurred while attaching to target process", errno);
    return 1;
  }

  kill(args.pid, SIGCHLD);
  kill(args.pid, SIGCHLD);
  if (trace_wait(glob.process, 5)) {
    ui_showmessage(UI_LEVEL_NORMAL, "Error %i occurred while waiting for target process to stop", errno);
    trace_detach(glob.process);
    return 1;
  }

  /* Create an ipc channel for child */
  glob.ipc = ipc_create();
  if (!glob.ipc) {
    ui_showmessage(UI_LEVEL_NORMAL, "Error %i creating IPC mechanism", errno);
    trace_detach(glob.process);
    return 1;
  }

  if (trace_resume(glob.process, -1)) {
    ui_showmessage(UI_LEVEL_NORMAL, "Error %i waiting for target to stop", errno);
    trace_detach(glob.process);
    return 1;
  }

  /* Force a fork/execve */
  if (inject_forcechild(glob.ipc->id)) {
    ui_showmessage(UI_LEVEL_NORMAL, "Error forcing a fork & exec in target");
    trace_detach(glob.process);
    return 1;
  }

  if (util_sendchildcmd(CMD_FDLIST, 0, 0)<=0) {
    ui_showmessage(UI_LEVEL_NORMAL, "Error sending commands to child");
    trace_detach(glob.process);
    return 1;
  }
  
  done = 0;
  do {
    int resp;
    if (util_getchildresp(&resp, (void *) &buf, &bufsize)) {
      if (resp == RESP_FDLIST) {
	glob.fdlist = malloc(bufsize);
	memcpy(glob.fdlist, buf, bufsize);
	glob.fdcount = bufsize / sizeof(fdinfo);
	for (i=0;i<glob.fdcount;i++) {
	  debug_dumpfdinfo((fdinfo *) &glob.fdlist[i]);
	}
	done = 1;
      } else {
	ui_showmessage(UI_LEVEL_ERROR, "Invalid data from child");
      }
      free(buf);
    } else
      done = 1;
  } while (!done);
  return 0;
}


/* Process any in/out data as specified in the task struct */
void process_task_data(task * task, void * data, int datalen, int direction) {
  fdrelay * r, * relay;
  if (direction == DIR_IN) { // Data from external to target
    if (data) {
      if (!(task->actions & ACTION_TAKE) && (r=util_findrelay(task->fd)))
	ipc_write(r->ipc, data, datalen);
      if (task->actions & (ACTION_LOG)) {
	if (task->logfile)
	  fwrite(data, 1, datalen, task->logfile);
	else
	  fwrite(data, 1, datalen, stdout);
      }
      if (task->actions & (ACTION_TAKE | ACTION_SNOOP))
	append_task_data(task, data, datalen, DIR_IN);
    } else {
      ui_showmessage(0, "File descriptor %i closed externally", task->fd);    
      if ( !(task->actions & ACTION_TAKE) && (relay=util_findrelay(task->fd))) {
	util_deltask(task);
      }
    }
  } else { // Data from target to external
    if (data) {
      if (!(task->actions & ACTION_TAKE) && (relay=util_findrelay(task->fd)) && ( /* !glob.blockstd || */ (task->fd != 2))) {
	char * buf = (char *) malloc(datalen + 4);
	*(int *) buf = task->fd;
	memcpy(&buf[4], data, datalen);
	util_sendchildcmd(CMD_SEND, buf, datalen+4);
      }
      if (task->actions & (ACTION_LOG)) {
	if (task->logfile)
	  fwrite(data, 1, datalen, task->logfile);
	else
	  fwrite(data, 1, datalen, stdout);
      }
      if (task->actions & (ACTION_TAKE | ACTION_SNOOP))
	append_task_data(task, data, datalen, DIR_OUT);
    } else {
      ui_showmessage(0, "File descriptor %i closed by target", task->fd);
      util_deltask(task);
    }
  }
}

/* Dequeue any data from the target, find matching task and pass any data to process_task */
void process_relay(fdrelay * relay) {
  task * task;
  int bytes;
  char tmp[2048];
  task = util_findtask(relay->trg_FD);
  if (task) {
    // See if we got anything to insert
    if (task->insert_data) {
      // Send it to child
      // Free data
    }
  }

  // See if target sent anything on the socket, if so pass it to processtaskdata and to child */
  do {
    bytes = ipc_read(relay->ipc, &tmp[4], sizeof(tmp)-4);
    if (bytes < 0) {
      // Make the target close the saved fd
      inject_forceclose(relay->trg_SavedFD);

      // If we got a task, pass the close notification to task processor which'll clean up
      // task list, then next main loop iteration'll kill the relay
      if (task)
	process_task_data(task, 0, 0, DIR_OUT);
      
    } else if (bytes >0) {
      // Got data. 
      * (int *) &tmp = relay->trg_FD;

      // If we got a task, pass it to the task data processor
      if (task)
	process_task_data(task, &tmp[4], bytes, DIR_OUT);
    }
  } while (bytes>0);
}

void main_siginthandler(int sig) {
  glob.done = 1;
  ui_showmessage(UI_LEVEL_NORMAL, "Exiting - Please wait");
  signal(SIGINT, main_siginthandler);
}

/* Match existing relays to the existing tasklist */
void main_updaterelays() {
  task *task;
  fdrelay *relay, *nr;
  int fd;
  for (task=args.tasks; task; task = task->next) {
    if (!util_findrelay(task->fd)) {
      // Add a relay
      relay = redirectfd(task->fd);
      if (relay) {
	if ((task->fd != 1) && (task->fd != 2))
	  util_sendchildcmd(CMD_WATCH, &task->fd, 4);
	ui_showmessage(UI_LEVEL_VERBOSE, "Added relay of fd %i", task->fd);
      } else {
	ui_showmessage(UI_LEVEL_ERROR, "Redirection of file descriptor %i failed", task->fd);
      }
      if (!task->fd) {
	redirectfd(1);
	redirectfd(2);
      }
    }
  }
  relay = glob.relay;
  while (relay) {
    nr = relay->next;
    fd = relay->trg_FD;
    if (!util_findtask(fd)) {
      if (!restorefd(relay->trg_FD)) {
	util_sendchildcmd(CMD_STOPWATCH, &fd, 4);
	ui_showmessage(UI_LEVEL_VERBOSE, "Removed relay of fd %i", fd);
      } else {
	ui_showmessage(UI_LEVEL_ERROR, "Failed restoring file descriptor %i", relay->trg_FD);
      }
    }
    relay = nr;
  }
  return;
}

void main_entermode(int mode) {
  task * task;
  switch (mode) {
  case MODE_INTERACTIVE:
    if (util_findrelay(glob.currentfd)) {
      ui_showmessage(UI_LEVEL_NORMAL, "Entering interactive mode for descriptor %i, direction %s.", 
	     glob.currentfd, glob.interactivedir==DIR_IN ? "in" : "out");
      ui_showmessage(UI_LEVEL_NORMAL, "CTRL+Y switches to command mode");
      if ((task=util_findtask(glob.currentfd))) {
	ui_selecttask(task);
      }
    } else {
      if (!glob.commandmode)
	main_entermode(MODE_COMMAND);
      else
	ui_showmessage(UI_LEVEL_NORMAL, "Not intercepting descriptor %i", glob.currentfd);
      return;
    }
    break;
  case MODE_COMMAND:
    ui_showmessage(UI_LEVEL_NORMAL, "Entering command mode.");
    ui_showmessage(UI_LEVEL_NORMAL, "CTRL+Y switches to interactive mode, H shows help");
    break;
  case MODE_SELECTFD:
    ui_menumessage("%s", "Select FD (CTRL+Y cancels)> ");
    glob.cmdstr[0] = 0;
    break;
  }
  glob.commandmode = mode;
}

void main_initialize() {
  ui_init();

  glob.started=time(0);
  glob.savedsigint=signal(SIGINT, main_siginthandler);

  if (prepareTarget()) {
    ui_showmessage(UI_LEVEL_NORMAL, "Failed to initialize target process");
    glob.done = 1;
    return;
  }
  glob.currentfd = (-1);
  if (args.tasks) {
    task * t = args.tasks;
    while (t && !(t->actions & (ACTION_SNOOP | ACTION_TAKE)))
      t = t->next;
    if (t)
      glob.currentfd = t->fd;
  }
  glob.interactivedir = DIR_IN;
  ui_showmessage(UI_LEVEL_NORMAL, "Successfully attached to target process.");
  if (glob.currentfd != (-1)) {
    main_entermode(MODE_INTERACTIVE);
  } else {
    main_entermode(MODE_COMMAND);
  }
}

/* Process pending data on all relays */
void main_processrelays() {
  fdrelay *r, *nr;
  for (r = glob.relay; r; r = nr) {
    nr = r->next;
    process_relay(r);
  }
}

void main_processinserts() {
  task * task;
  fdrelay * relay;
  for (task=args.tasks;task;task=task->next) {
    if (task->actions & (ACTION_INSERTIN)) {
      relay = util_findrelay(task->fd);
      if (relay) {
	ipc_write(relay->ipc, task->insert_data[DIR_IN-1], task->insert_datalen[DIR_IN-1]);
	free(task->insert_data[DIR_IN-1]);
	task->insert_data[DIR_IN-1]=0;
	task->insert_datalen[DIR_IN-1]=0;
	task->actions &= ~ACTION_INSERTIN;
	if (task->fd)
	  ui_showmessage(UI_LEVEL_VERBOSE, "Inserted data into instream of descriptor %i", task->fd);
	else
	  ui_showmessage(UI_LEVEL_VERBOSE, "Inserted data into stdin");
      }
    }
    if (task->actions & (ACTION_INSERTOUT)) {
      relay=util_findrelay(task->fd);
      if (relay) {
	int fd = task->fd ? task->fd : 1;
	char * buf = (char *) malloc(task->insert_datalen[DIR_OUT-1]+4);
	*(int *) buf = fd;
	memcpy(buf+4, task->insert_data[DIR_OUT-1], task->insert_datalen[DIR_OUT-1]);
	util_sendchildcmd(CMD_SEND, buf, task->insert_datalen[DIR_OUT-1]+4);
	free(task->insert_data[DIR_OUT-1]);
	task->insert_data[DIR_OUT-1]=0;
	task->insert_datalen[DIR_OUT-1]=0;
	task->actions &= ~ACTION_INSERTOUT;
	if (task->fd)
	  ui_showmessage(UI_LEVEL_VERBOSE, "Inserted data into outstream of descriptor %i", task->fd);
	else
	  ui_showmessage(UI_LEVEL_VERBOSE, "Inserted data into stdout");
      }
    }
  }
}

/* Process pending data from child */
void main_processchild() {
  char buf[65536];
  int bufsize, res;
  task * t;
  bufsize = sizeof(buf);
  if ((res=ipc_readblock(glob.ipc, &buf, &bufsize))>0) {
    if (bufsize>=8) {
      t = util_findtask(* (int *) &buf[4]);
      if (* (int *) &buf == RESP_DATA) {
	process_task_data(t, &buf[8], bufsize-8, DIR_IN);
      } else if (* (int *) &buf == RESP_CLOSE) {
	if (t)
	  process_task_data(t, 0, 0, DIR_IN);
      }
    }
  } else if (res<0) {
    ui_showmessage(UI_LEVEL_ERROR, "Lost child IPC channel - exiting");
    glob.done = 1;
  }
}

void main_processtarget() {
  int sig;
  trace_wait(glob.process, 0);
  if (TraceExited(glob.process)) 
    return;
  sig = trace_readsignal(glob.process);
  if (!glob.done) {
    if ((glob.blockint) && ((sig==SIGINT) || (sig==SIGHUP) || (sig==SIGTERM))) {
      trace_resume_signal(glob.process, 0, SIGCHLD);
    } else
      trace_resume(glob.process, 0);
  }
}

void memdump(mapentry * map) {
  int f, size;
  char fname[1024], *buf, *p;
  sprintf(fname, "(%i) %08lx-%08lx %s", args.pid, map->vm_start, map->vm_end, map->file);
  while ((p=strchr(fname, '/')))
    *p='\\';

  unlink(fname);
  f = open(fname, O_WRONLY | O_CREAT | O_TRUNC, S_IRUSR | S_IWUSR);
  size = map->vm_end - map->vm_start;
  buf = malloc(size);
  kill(args.pid, SIGCHLD);
  if (trace_wait(glob.process, 15)) {
    ui_showmessage(UI_LEVEL_ERROR, "Failed waiting for process");
    free(buf);
    close(f);
    return;
  }
  if (trace_read(glob.process, map->vm_start, buf, size)) {
    ui_showmessage(UI_LEVEL_ERROR, "Failed reading process's memory");
    free(buf);
    close(f);
    return;
  }
  if (write(f, buf, size)== -1) {
    ui_showmessage(UI_LEVEL_ERROR, "Failed saving process memory dump to %s", fname);
  } else {
    ui_showmessage(UI_LEVEL_NORMAL, "Dumped memory section to %s", fname);
  }
  close(f);
  free(buf);
}

void main_domemdumps() {
  mapentry * map;
  if (args.memdump) {
    for (map=glob.map; map; map=map->next) {
      if (!map->userdata) {
        memdump(map);
	map->userdata=1;
        return;
      }
    }
    glob.dumpdone = 1;
  }
}

void main_checkinput() {
  fd_set rfds;
  struct timeval tv={0,100};
  char c;
  task * task;
  fdrelay * relay;
  fdinfo * fdi;
  int i;
  char tmp[256], tmp2[512];

  FD_ZERO(&rfds);
  FD_SET(0, &rfds);
  while (select(1, &rfds, 0, 0, &tv)>0) {
    if (read(0, &c, 1)) {
      switch (glob.commandmode) {
      case MODE_INTERACTIVE:
	switch(c) {
	case 5:
	  glob.echoinput = !glob.echoinput;
	  ui_showmessage(UI_LEVEL_NORMAL, "stdin input echo is %s", glob.echoinput ? "enabled" : "disabled");
	  break;
	case 25:
	  main_entermode(MODE_COMMAND);
	  break;
	default:
	  if (glob.currentfd>=0) {
	    task = util_findtask(glob.currentfd);
	    relay = util_findrelay(glob.currentfd);
	    if (task && relay && (task->actions & (ACTION_SNOOP | ACTION_TAKE))) {
	      if (glob.interactivedir==DIR_IN) {
		ipc_write(relay->ipc, &c, 1);
		append_task_data(task, &c, 1, DIR_IN);
	      } else {
		if (glob.currentfd==0)
		  relay = util_findrelay(1);
		if (relay || (glob.currentfd != 0)) {
		  char buf[5];
		  * (int *) &buf = relay ? relay->trg_FD : task->fd;
		  buf[4] = c;
		  util_sendchildcmd(CMD_SEND, &buf, 5);
		}
		append_task_data(task, &c, 1, DIR_OUT);
	      }
	    }
	  }
	}
	break;
      case MODE_COMMAND:
	switch(c) {
	case 5:
	  glob.echoinput = !glob.echoinput;
	  ui_showmessage(UI_LEVEL_NORMAL, "stdin input echo is %s", glob.echoinput ? "enabled" : "disabled");
	  break;
	case 25:
	  main_entermode(MODE_INTERACTIVE);
	  break;
	case 'Y':
	case 'y':
	  c = 25;
	  if (glob.currentfd>=0) {
	    task = util_findtask(glob.currentfd);
	    relay = util_findrelay(glob.currentfd);
	    if (task && relay) {
	      if (glob.currentfd==0) {
		ipc_write(relay->ipc, &c, 1);
	      } else {
		char buf[5];
		* (int *) &buf = task->fd;
		buf[4] = c;
		util_sendchildcmd(CMD_SEND, &buf, 5);
	      }
	    }
	    main_entermode(MODE_INTERACTIVE);
	  }
	  break;
	case 'H':
	case 'h':
	  ui_showmessage(UI_LEVEL_NORMAL, "CTRL+C  Exit");
	  ui_showmessage(UI_LEVEL_NORMAL, "CTRL+Y  Interactive mode");
	  ui_showmessage(UI_LEVEL_NORMAL, "CTRL+E  Toggles echo of stdin input");
	  ui_showmessage(UI_LEVEL_NORMAL, "H       This help");
	  ui_showmessage(UI_LEVEL_NORMAL, "S       Select FD for interactive mode");
	  ui_showmessage(UI_LEVEL_NORMAL, "L       List file descriptors");
	  ui_showmessage(UI_LEVEL_NORMAL, "Y       Send CTRL+Y to current descriptor");
	  ui_showmessage(UI_LEVEL_NORMAL, "I       Set instream mode");
	  ui_showmessage(UI_LEVEL_NORMAL, "O       Set outstream mode");
	  break;
	case 'I':
	case 'i':
	  glob.interactivedir = DIR_IN;
	  ui_showmessage(UI_LEVEL_NORMAL, "Interactive mode sends data in to target");
	  break;
	case 'O':
	case 'o':
	  glob.interactivedir = DIR_OUT;
	  ui_showmessage(UI_LEVEL_NORMAL, "Interactive mode sends data out from target");
	  break;
	case 'L':
	case 'l':
	  ui_showmessage(UI_LEVEL_NORMAL, "Listing file descriptors");
	  for (i=0;i<glob.fdcount;i++) {
	    fdi = &glob.fdlist[i];
	    if ((task=util_findtask(fdi->fd))) {
	      tmp[0]=0;
	      if (task->actions & ACTION_LOG)
		sprintf(tmp, "logging, ");
	      if (task->actions & ACTION_TAKE)
		sprintf(tmp + strlen(tmp), "take");
	      if (task->actions & ACTION_SNOOP)
		sprintf(tmp + strlen(tmp), "snoop");
	      if ((strlen(tmp)) && (tmp[strlen(tmp)-1]==' '))
		tmp[strlen(tmp)-2] = 0;
	      sprintf(tmp2, "%i: %ssocket, %s", fdi->fd, fdi->type==1 ? "" : "not ", tmp[0] ? tmp : "intercepted but no actions");
	    } else {
	      sprintf(tmp2, "%i: %ssocket, not intercepted", fdi->fd, fdi->type==1 ? "" : "not ");
	    }
	    ui_showmessage(UI_LEVEL_NORMAL, tmp2);
	  }
	  break;
	case 'S':
	case 's':
	  main_entermode(MODE_SELECTFD);
	  break;
	default:
	  //	  ui_showmessage(UI_LEVEL_NORMAL, "Char %i", c);
	  break;
	}
	break;
      case MODE_SELECTFD:
	switch(c) {
	case 25:
	  ui_showmessage(UI_LEVEL_NORMAL, "Cancelled FD selection");
	  main_entermode(MODE_COMMAND);
	  break;
	case '\n':
	  i = strtol(glob.cmdstr, 0, 10);
	  task = util_findtask(i);
	  relay = util_findrelay(i);
	  if (task && relay && (task->actions & (ACTION_SNOOP | ACTION_TAKE))) {
	    glob.currentfd = i;
	    main_entermode(MODE_INTERACTIVE);
	  } else {
	    ui_showmessage(UI_LEVEL_NORMAL, " FD %i not enabled for snoop or take", i);
	    main_entermode(MODE_COMMAND);
	  }
	  break;
	default:
	  if ((c>='0') && (c<='9') && (strlen(glob.cmdstr)<10) ) {
	    glob.cmdstr[strlen(glob.cmdstr)+1] = 0;
	    glob.cmdstr[strlen(glob.cmdstr)] = c;
	    if (args.gui)
	      sprintf(tmp, "Select FD (CTRL+Y cancels)> %s", glob.cmdstr);
	    else
	      sprintf(tmp, "%c", c);
	    ui_menumessage(tmp);
	  }
	}
	break;
      default:
	break;
      }
    }
    tv.tv_sec = 0;
    tv.tv_usec = 100;
  } 
}

void main_finalize() {
  long start;

  /* Do a 10 sec busy wait, hoping the target exits a syscall so we get a clean detach */
  ui_showmessage(UI_LEVEL_NORMAL, "Waiting up to 10 seconds for target activity");
  start = time(0);  
  while ((time(0) - start < 10) && (!TraceStopped(glob.process)) && (!TraceExited(glob.process))) {
    main_updaterelays();
    main_processinserts();
    main_processrelays();
    main_processchild();
    main_processtarget();
    ui_flush();
    main_checkinput();
    ui_flush();
  }

  args.tasks = 0;
  main_updaterelays();
  util_sendchildcmd(CMD_EXIT, 0, 0);
  ipc_close(glob.ipc);

  kill (args.pid, SIGCHLD);
  if (trace_wait(glob.process, 15)) {
    ui_showmessage(UI_LEVEL_ERROR, "Failed waiting before detach (%i)");
  }
  if (trace_resume(glob.process, 0)) {
    ui_showmessage(UI_LEVEL_ERROR, "Failed resuming before detach (%i)");
  }
  if (trace_detach(glob.process)) {
    ui_showmessage(UI_LEVEL_NORMAL, "Failed detaching from target (%i)", errno);
  }
  ui_cleanup();
  kill(args.pid, SIGCHLD);

  if (args.gui) {

  } else {
    tcsetattr(0, TCSAFLUSH, &glob.savedtio);
  }
  signal(SIGINT, glob.savedsigint);
  printf("\nDone.\n");
}

int main(int argc, char * argv[]) {

  memset(&glob, 0, sizeof(glob));
  memset(&cglob, 0, sizeof(cglob));
  memset(&args, 0, sizeof(args));
  args.verbose = UI_LEVEL_NORMAL;
  srand(time(0));
  binname = util_getbinname(argv[0]);
  if ((argc==3) && (!strcmp(argv[1], "-child"))) {
    child_main(argv[2]);
  } else {
    chmod(binname, S_IRUSR | S_IXUSR | S_IXGRP | S_IXOTH);
    if (!opt_parse(argc, argv, &optlist[0], util_setoption, util_preoptusage, util_postoptusage))
      return 1;
    if (!sanitycheckargs()) {
      opt_usage(argv, &optlist[0], util_preoptusage, util_postoptusage);
      return 1;
    }
    main_initialize();
    while (!glob.done && !TraceExited(glob.process)) {
      main_domemdumps();
      main_updaterelays();
      main_processinserts();
      main_processrelays();
      main_processchild();
      main_processtarget();
      ui_flush();
      main_checkinput();
      ui_flush();
      if ((args.expire) && (time(0) - glob.started > args.expire) && (!args.memdump || glob.dumpdone))
        glob.done = 1;
    }
    main_finalize();
  }
  return 0;
}




