/*

This file is (C) Copyright 2002 Erik S. Johansen <einride@einride.org>, all 
rights reserved.

If you work for or represent an organisation, agency or corporation that is 
wholly or partially funded by a government -OR- you are in any way associated 
with the company Cyber Operations, the following licence applies.

You may not use this file or any derivatives of it for any purpose whatsoever,
including reading beyond this comment, redistribution or compiling, without 
explicit written permission from the copyright holder. 

If you do not work for or represent an organisation, agency or corporation 
that is wholly or partially funded by a government -AND- you are not in any 
way associated with the company Cyber Operations, the GNU General Public 
License, as described in the accompanying file GPL, applies.

*/
#include "util.h"
#include "ui.h"
#include "options.h"

char * binname = 0;
struct app_arg_struct args;
struct app_glob_struct glob;
struct child_glob_struct cglob;

/*****************************************************************
 * Available option list as options.c:opt_parse wants them
 *****************************************************************/
struct option optlist[] = {
  {"pid",        'p', "<pid>",                            "Process ID of target."},
  {"log",        'l', "std|<fd> filename",                "Logs the specified fd/std* to file"},
  {"watch",      'w', "std|<fd>",                         "Dumps the specified fd/std* to stdout"},
  {"insertin",   'i', "std|<fd> data|@file",              "Inserts data into the instream of the fd."},
  {"insertout",  'o', "std|<fd> data|@file",              "Inserts data into the outstream of the fd."},
  {"expire",     'e', "<seconds>",                        "Exits the hijacker after specified interval."},
  {"verbose",    'v', "",                                 "Enables verbose descriptions of what happens."},
  {"take",       't', "std|<fd>",                         "Takes the specified fds/std*."},
  {"snoop",      's', "std|<fd>",                         "Snoops the specified fds/std*."},
  {"memdump",    'm', "",                                 "Dump the process's memory"},
  /*  {"gui",        'g', "",                                 "Enters GUI mode"}, */
  {NULL, 0, NULL, NULL}
};



/*****************************************************************
 * Build a linked list of mapentry* by reading the 
 * specified pid's /proc/pid/maps
 *****************************************************************/
mapentry * util_makemapentry(unsigned int pid) {
  int f, lpos, res;
  char fname[256], *p, pstr[256];
  char line[4096];
  mapentry * map = 0, *cmap = 0;
  sprintf(fname, "/proc/%i/maps", pid);
  f = open(fname, O_RDONLY);
  if (f<0)
    return 0;
  do {
    lpos=0;
    while ( ((res=read(f, &line[lpos], 1))==1) && (line[lpos]!='\n'))
      lpos++;
    memset(&line[lpos], 0, sizeof(line)-lpos);
    line[lpos] = 0;
    if (line[0]) {
      if (cmap) {
	cmap->next = (mapentry *) malloc (sizeof(mapentry));
	cmap=cmap->next;
      } else {
	map = (mapentry *) malloc (sizeof(mapentry));
	cmap = map;
      }
      memset(cmap, 0, sizeof(mapentry));

      sscanf(line, "%08lx-%08lx %s %08lx %s %lu %s", 
	     &cmap->vm_start, &cmap->vm_end, pstr, &cmap->vm_pgoff, 
	     cmap->devname, &cmap->inode, cmap->file);
      p=pstr;
      cmap->vm_read = (*p++=='r');
      cmap->vm_write = (*p++=='w');
      cmap->vm_exec = (*p++=='x');
      cmap->vm_shared = (*p++!='p');
    }
  } while (res==1);
  close(f);
  return map;
}

/*****************************************************************
 * Read the specified pid's /proc/pid/cmdline and guess what the
 * primary binary name is
 *****************************************************************/
int util_getprocessbinname(unsigned int pid, int removepath, char * buf, int bufsize) {
  int f;
  char line[4096], fname[4096], *p;
  sprintf(fname, "/proc/%i/cmdline", pid);
  f=open(fname, O_RDONLY);
  if (f<0)
    return 0;
  memset(line, 0, sizeof(line));
  if (read(f, line, sizeof(line)-1)>=1) {
    p = strchr(line, ' ');
    if (p)
      *p=0;
    if (removepath) {
      p=strrchr(line, '/');
      if (!p)
	p=line;
    } else
      p=line;
    if (bufsize > strlen(p)) {
      strcpy(buf, p);
      return 1;
    } else
      return 0;
  } else {
    close(f);
    return 0;
  }
}


/*****************************************************************
 * arg0 should be argv[0] as passed to main. Build a fully 
 * qualified path & filename, optionally using pwd. malloc and 
 * return a str.
 *****************************************************************/
char * util_getbinname(char * arg0) {
  char tmp[8192], *res, *p, *p2, *arg, argv0[8192];
  strcpy(argv0, arg0);
  arg = argv0;
  if (arg[0]=='/') {
    res = (char *) malloc(strlen(arg)+1);
    strcpy(res, arg);
    return res;
  }
  getcwd(tmp, sizeof(tmp));
  if (tmp[strlen(tmp)-1]!='/') 
    strcat(tmp, "/");
  while ((p=strchr(arg, '/'))) {
    *p++ = 0;
    if (!strcmp(arg, "..")) {
      tmp[strlen(tmp)-1] = 0;
      p2 = strrchr(tmp, '/');
      if (p2) {
	p2++;
	*p2 = 0;
      }
    } else if (strcmp(arg, ".")) {
      strcat(tmp, arg);
      strcat(tmp, "/");
    }
    arg = p;
  }
  strcat(tmp, arg);
  res = malloc(strlen(tmp)+1);
  strcpy(res, tmp);
  return res;
}


/*****************************************************************
 * Add a task to glob.task linked list
 *****************************************************************/
task * util_addtask() {
  task *t, *t2;
  t = (task *) malloc(sizeof(task));
  memset(t, 0, sizeof(task));
  if (args.tasks) {
    t2 = args.tasks;
    while (t2->next)
      t2 = t2->next;
    t2->next = t;
  } else {
    args.tasks = t;
  }
  return t;
}

/*****************************************************************
 * Remove a task from glob.task linked list
 *****************************************************************/
void util_deltask(task * tsk) {
  task * t;
  if (tsk==args.tasks) 
    args.tasks = tsk->next;
  else {
    for (t=args.tasks; t; t = t->next)
      if (t->next==tsk)
	break;
    if (t)
      t->next = tsk->next;
  }
  if (tsk->insert_data[0])
    free(tsk->insert_data[0]);
  if (tsk->insert_data[1])
    free(tsk->insert_data[1]);
  if (tsk->logfile)
    fclose(tsk->logfile);
  memset(tsk, 0, sizeof(task));
  free(tsk);
}
/*****************************************************************
 * Find task struct in glob.task linked list for a given fd
 *****************************************************************/
task * util_findtask (int fd) {
  task * t = args.tasks;
  if ((fd==1) || (fd==2))
    fd = 0;
  while (t && (t->fd != fd))
    t = t->next;
  return t;
}


/*****************************************************************
 * add a relay to glob.relay linked list
 *****************************************************************/
fdrelay * util_addrelay() {
  fdrelay * relay, *r;
  relay = (fdrelay *) malloc(sizeof(fdrelay));
  memset(relay, 0, sizeof(fdrelay));
  if (glob.relay) {
    r=glob.relay;
    while (r->next)
      r = r->next;
    r->next = relay;
  } else
    glob.relay = relay;

  /* Create a listening unix socket */
  relay->ipc = ipc_create();
  if (!relay->ipc) {
    ui_showmessage(UI_LEVEL_ERROR, "Failed creating IPC channel for redirection of file descriptor");
    return 0;
  }
  return relay;
}

/*****************************************************************
 * remove a relay from relay linked list
 *****************************************************************/
void util_delrelay (fdrelay * rel) {
  fdrelay * r;
  if (rel == glob.relay) 
    glob.relay = rel->next;
  else {
    r = glob.relay;
    while (r && (r->next!=rel))
      r = r->next;
    if (r)
      r->next = rel->next;
  }
  ipc_close(rel->ipc);
  free(rel);
}

/*****************************************************************
 * Find relay struct for a given fd
 *****************************************************************/
fdrelay * util_findrelay (int fd) {
  fdrelay * r = glob.relay;
  while (r && (r->trg_FD != fd))
    r = r->next;
  return r;
}

/*****************************************************************
 * Fixup any \quotes and set newlen to actual data length
 *****************************************************************/
int util_parsequotes(char * src, int * len) {
  char * trg = (char *) malloc((*len) * 2);
  char * p, tmp[256];
  int spos=0, tpos=0, i;
  while (spos < *len) {
    if (src[spos]=='\\') {
      spos++;

      /* \OOO \xHH \n \t \r */
      
      strncpy(tmp, (char *) &src[spos], 4);
      tmp[3] = 0;
      i = strtol(tmp, &p, 8);
      if ((*p==0) && (i>=0) && (i<=255)) {
	trg[tpos++] = i;
	spos += 3;
	continue;
      }
      if (tmp[0]=='x') {
	i = strtol((char *) &tmp[1], &p, 16);
	if ((*p==0) && (i>=0) && (i<=255)) {
	  trg[tpos++] = i;
	  spos += 3;
	  continue;
	}
      }
      switch (tmp[0]) {
      case 'n':
	trg[tpos++]='\n';
	spos++;
	break;
      case 't':
	trg[tpos++]='\t';
	spos++;
	break;
      case 'r':
	trg[tpos++]='\r';
	spos++;
	break;
      default:
	trg[tpos++]=tmp[0];
      }
    } else
      trg[tpos++] = src[spos++];
  }
  trg[tpos] = 0;
  memcpy(src, trg, tpos+1);
  *len = tpos;
  free(trg);
  return 1;
}

/*****************************************************************
 * Add one or more log specifications to task list
 *****************************************************************/
int util_addtask_log(char * what, char * file) {
  char *p="";
  int fd;
  task * t;
  FILE * f;
  if (!strcmp(what, "std"))
    fd = 0;
  else
    fd = strtol(what, &p, 10);
  if ((fd==1) || (fd==2)) 
    fd = 0;
  if ((*p!=0) || (fd<0)) {
    ui_showmessage(UI_LEVEL_ERROR, "Invalid file descriptor %s", what);
    return 0;
  }
  t=util_findtask(fd);
  if (t && (t->actions & ACTION_LOG)) {
    ui_showmessage(UI_LEVEL_ERROR, "Can't log file descriptor %i twice", fd);
    return 0;
  }
  if (file) {
    f = fopen(file, "a");
    if (!f) {
      ui_showmessage(UI_LEVEL_ERROR, "Can't open %s for writing (%i)", file, errno);
      return 0;
    }
  } else
    f = 0;
  if (!t)
    t = util_addtask();
  t->fd = fd;
  t->actions |= ACTION_LOG;
  t->logfile = f;
  return 1;
}

/*****************************************************************
 * Add one or more insert specifications to task list
 *****************************************************************/
int util_addtask_insert(char * what, char * data, int direction) {
  char *p = "", *buf;
  int fd, act, datalen;
  task * t;
  
  switch(direction) {
  case DIR_IN:
    act=ACTION_INSERTIN;
    break;
  case DIR_OUT:
    act=ACTION_INSERTOUT;
    break;
  default:
    return 0;
  }
  if (!strcmp(what, "std"))
    fd = 0;
  else
    fd = strtol(what, &p, 10);
  if ((fd==1) || (fd==2))
    fd = 0;
  if ((*p!=0) || (fd<0)) {
    ui_showmessage(UI_LEVEL_ERROR, "Invalid file descriptor %s", what);
    return 0;
  }
  t=util_findtask(fd);
  if (t && (t->actions & act)) {
    ui_showmessage(UI_LEVEL_ERROR, "Can't insert into descriptor %i twice", fd);
    return 0;
  }
  
  if (data[0]=='@') {
    FILE * f = fopen((char*) &data[1], "r");
    if (!f)
      return 0;
    fseek(f, SEEK_END, 0);
    datalen=ftell(f);
    fseek(f, SEEK_SET, 0);
    buf = malloc(datalen);
    fread(buf, 1, datalen, f);
    fclose(f);
  } else {
    datalen = strlen(data);
    if (!util_parsequotes(data, &datalen))
      return 0;
    buf = (char *) malloc(datalen);
    memcpy(buf, data, datalen);
  }

  if (!t)
    t = util_addtask();
  t->fd = fd;
  t->actions |= act;
  t->insert_data[direction-1] = malloc(datalen);
  memcpy(t->insert_data[direction-1], buf, datalen);
  t->insert_datalen[direction-1] = datalen;
  free(buf);
  return 1;
}

/*****************************************************************
 * Add one or more take specifications to task list
 *****************************************************************/
int util_addtask_take(char * what) {
  char *p = "";
  int fd;
  task * t;

  if (!strcmp(what, "std"))
    fd = 0;
  else
    fd = strtol(what, &p, 10);
  if ((fd==1) || (fd==2))
    fd = 0;
  if ((*p!=0) || (fd<0)) {
    ui_showmessage(UI_LEVEL_ERROR, "Invalid file descriptor %s", what);
    return 0;
  }
  t=util_findtask(fd);
  if (t && (t->actions & (ACTION_TAKE | ACTION_SNOOP))) {
    ui_showmessage(UI_LEVEL_ERROR, "Can only use one of watch, take, snoop (fd %i)", fd);
    return 0;
  }
  if (!t)
    t = util_addtask();
  t->fd = fd;
  t->actions |= ACTION_TAKE;
  return 1;
}

/*****************************************************************
 * Add one or more snoop specifications to task list
 *****************************************************************/
int util_addtask_snoop(char * what) {
  char *p = "";
  int fd;
  task * t;

  if (!strcmp(what, "std"))
    fd = 0;
  else
    fd = strtol(what, &p, 10);
  if ((fd==1) || (fd==2))
    fd = 0;
  if ((*p!=0) || (fd<0)) {
    ui_showmessage(UI_LEVEL_ERROR, "Invalid file descriptor %s", what);
    return 0;
  }
  t=util_findtask(fd);
  if (t && (t->actions & (ACTION_TAKE | ACTION_SNOOP))) {
    ui_showmessage(UI_LEVEL_ERROR, "Can only use one of watch, take, snoop (fd %i)", fd);
    return 0;
  }
  if (!t)
    t = util_addtask();
  t->fd = fd;
  t->actions |= ACTION_SNOOP;
  return 1;
}


/* Called by opt_parse to parse arguments to an option */
int util_setoption(int ndx, int argc, char * argv[]) {
  char * p;
  int res;
  switch (ndx) {
  case 0: /* pid */
    if (args.pid || (argc!=1))
      return 0;
    args.pid = strtol(argv[0], &p, 10);
    if ((args.pid <= 0) || (*p!=0))
      return 0;
    else
      return 1;
  case 1: /* log */
    if ( (argc<1) || (argc>2) )
      return 0;
    else
      return util_addtask_log(argv[0], (argc==2) ? argv[1] : NULL);
  case 2: /* watch */
    if ((argc!=1))
      return 0;
    else
      return util_addtask_log(argv[0], NULL);
  case 3: /* insertin */
    if (argc!=2)
      return 0;
    return util_addtask_insert(argv[0], argv[1], DIR_IN);
  case 4: /* insertout */
    if (argc!=2)
      return 0;
    return util_addtask_insert(argv[0], argv[1], DIR_OUT);
  case 5: /* expire */
    if (args.expire || (argc!=1))
      return 0;
    args.expire = strtol(argv[0], &p, 10);
    if ((args.expire <= 0) || (*p!=0))
      return 0;
    return 1;
  case 6: /* verbose */
    args.verbose++;
    return 1;
  case 7: /* take */
    if (argc != 1)
      return 0;
    res = util_addtask_take(argv[0]);
    return res;
  case 8: /* snoop */
    if (argc != 1)
      return 0;
    res = util_addtask_snoop(argv[0]);
    return res;
  case 9: /* memdump */
    if (args.memdump || argc)
      return 0;
    args.memdump = 1;
    return 1;
  case 10: /* gui */
    if (args.gui || argc)
      return 0;
    args.gui = 1;
    return 1;
  default:
    return 0;
  }
}

void util_preoptusage( char * bin) {
  fprintf(stderr, "Usage: %s <Options>\nAvailable options:\n", bin);
  return;
}

void util_postoptusage(char * bin) {
  fprintf(stderr, "Only one of --take and --snoop may be specified.\n");
  /*  fprintf(stderr, "If --gui is specified, settings may be changed interactively.\n"); */
  return;
}


int util_sendchildcmd(int cmd, void * data, int datalen) {
  int i, res;
  char * buf;
  if (data && datalen) {
    buf = malloc(datalen+4);
    * (int *) buf = cmd;
    memcpy(&buf[4], data, datalen);
  } else {
    buf = (char *) &cmd;
    datalen = 0;
  }
  for (i=0;i<30;i++) {
    res = ipc_sendblock(glob.ipc, buf, datalen + 4);
    if (res>0) {
      if (datalen>0)
	free(buf);      
      return res;
    } else if (res<0) {
      if (datalen>0)
	free(buf);
      return 0;
    }
    usleep(100);
  }
  if (datalen>0)
    free(buf);
  return 0;
}

int util_getchildresp(int * resp, void ** data, int * datasize) {
  void * buf = 0;
  int bufsize = 0, buflen, res, i;
  *data = 0;
  *datasize = 0;
  for (i=0;i<10;i++) {
    do {
      bufsize += 1024;
      buf = realloc(buf, bufsize);
      buflen = bufsize;
      res = ipc_readblock(glob.ipc, buf, &buflen);
    } while (!res && buflen);
    if (res>0) {
      *datasize = buflen - 4;
      *resp = *(int *) buf;
      memmove(buf, buf + 4, buflen - 4);
      (*data) = buf;
      return 1;
    } else if (!res)
      usleep(500);
    else 
      return 0;
  };
  if (buf)
    free(buf);
  return 0;
}

void util_cyclicbufferappend(char * buffer, int buffersize, int * buffirst, int * bufnext,
			    char * data, int datalen) {
  int i, bpos;
  if (datalen > buffersize) {
    data += (datalen - buffersize);
    datalen = buffersize;
  }
  for (i=0, bpos=(*bufnext);i<datalen;i++) {
    if (bpos==buffersize)
      bpos=0;
    buffer[bpos++] = data[i];
    if ((*buffirst) == bpos)
      *buffirst = bpos + 1;
    if (*buffirst == buffersize)
      *buffirst = 0;
  }
  *bufnext = bpos;
}



